#extractMoboCfgInfo.awk
#
#This script parses the configuration file for a motherboard and extracts all
# of the "config." and "misc." tags.  Note when this is called, the config
# file must be known to match the motherboard.

#Set our field separator to "="
BEGIN {
    FS = "="
    mbModel="none specified" 
    mbNmiSmiDate="none specified" 
    mbNmiSmiRev="none specified" 
    mbNmiSmiGetDate="none specified" 
    mbNmiSmiGetRev="none specified" 
    mbNumPorts=0
    mbNumSlots=0 
    mbNumMisc=0
    mbNumRenameIf=0
    mbNumHideIf=0
    #Some tags are mandatory, we'll make sure we set these to 1 when we find
    # them in the cfg file.
    readNumPorts=0
    readNumSlots=0
    slotsCounted=0
    portsCounted=0    

    res = 1
    logPrefix = "mobomap: " 
    #set our success code to 0 (error) in case the script dies for some reason
    printf( "res=0\n" )
}

#Set our output variables
END { 
    if (readNumPorts == 0) 
    {
        system ("logger -p local0.error \"" logPrefix "(" argFileName ") The number of ports is a mandatory tag but was not found\"")
        res = 2
    }
    if (readNumSlots == 0) 
    {
        system ("logger -p local0.error \"" logPrefix "(" argFileName ") The number of slots is a mandatory tag but was not found\"")
        res = 2
    }

    printf( "mbModel=\"%s\"\n", mbModel ) 
    printf( "mbNmiSmiDate=\"%s\"\n", mbNmiSmiDate ) 
    printf( "mbNmiSmiRev=\"%s\"\n", mbNmiSmiRev ) 
    printf( "mbNmiSmiGetDate=\"%s\"\n", mbNmiSmiGetDate ) 
    printf( "mbNmiSmiGetRev=\"%s\"\n", mbNmiSmiGetRev ) 
    
    printf( "mbNumPorts=\"%s\"\n", mbNumPorts ) 
    for (i=1; i<=mbNumPorts; ++i)
    {
        if (mbPort[i] == "") 
        {
            #default unspecified ports to their number
            mbPort[i] = i
            #need to increment since defaulted ports count as a port found
            ++portsCounted
        }
        #Check each port for uniqueness
        if( uniquePorts[mbPort[i]] == 1 )
        {
            system ("logger -p local0.error \"" logPrefix "(" argFileName ") Duplicate reference to motherboard port (mobo.ports.port" i "=" mbPort[i] ")\"")
            res = 2
        }
        else
        {
            uniquePorts[mbPort[i]] = 1 
        }
        printf( "mbPort[%s]=\"%s\"\n", i, mbPort[i] ) 
    }
    if (portsCounted > mbNumPorts) 
    {
        system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Too many ports specified ("portsCounted"); config specifies "mbNumPorts"\"")
    }
    
    printf( "mbNumSlots=\"%s\"\n", mbNumSlots ) 
    for (i=1; i<=mbNumSlots; ++i)
    {
        if (mbSlot[i] == "") 
        {
            system ("logger -p local0.error \"" logPrefix "(" argFileName ") No PCI select found for slot "i"\"")
            res = 2
        }
        printf( "mbSlot[%s]=\"%s\"\n", i, mbSlot[i] ) 
        printf( "mbSlotRev[%s]=\"%s\"\n", i, mbSlotRev[i] ) 
    }
    if (slotsCounted > mbNumSlots) 
    {
        system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Too many slots specified ("slotsCounted"); config specifies "mbNumSlots"\"")
    }

    printf( "mbNumPortNames=\"%s\"\n", mbNumPortNames ) 
    for (i=0; i<mbNumPortNames; ++i)
    {
        printf( "mbPortNames[%s]=\"%s\"\n", i, mbPortNames[i] ) 
    }

    printf( "mbNumMisc=\"%s\"\n", mbNumMisc ) 
    for (i=0; i<mbNumMisc; ++i)
    {
        printf( "mbMisc[%s]=\"%s\"\n", i, mbMisc[i] ) 
    }

    printf( "mbNumRenameIf=\"%s\"\n", mbNumRenameIf )
    for (i=0; i<mbNumRenameIf; ++i)
    {
        printf( "mbRenameIf[%s]=\"%s\"\n", i, mbRenameIf[i] )
    }

    printf( "mbNumHideIf=\"%s\"\n", mbNumHideIf )
    for (i=0; i<mbNumHideIf; ++i)
    {
        printf( "mbHideIf[%s]=\"%s\"\n", i, mbHideIf[i] )
    }

    #set our success code to 1
    printf( "res=\"" res "\"\n" )
}                    

function CheckForDupes( field1, field2 )
{
    if (tags[field1] == 1) 
    {
        system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Duplicate tag found (" field1 "=" field2 ")\"")
    }
    tags[field1]=1
    return 0
}

#____________________________________________________________________
# This script expects the following input parameters:
#   argPciconfRes - The results of running "pciconf -l -v"
#   argSysinfoRes - The results of running "sysinfo"
#   argFileName - Name of the file being parsed
#
# This script parses specified config file and sets the following
# variables based on the configuration tags found:
#   mbModel - string of all of the motherboards described by this file
#   mbNmiSmiDate - string to indicate the min NMI/SMI support bios date
#   mbNmiSmiRev - string to indicate the min NMI/SMI support bios rev
#   mbNumPorts - number of ports on the motherboard
#   mbPort[] - array of pci device names mapping physical to named ports
#   mbNumSlots - number of pci expansion slots
#   mbSlotSel[] - array of pci selects to use
#   mbSlotRev[] - array of bool to indicate whether slot is reversed
#   mbMiscTags[] - array of miscellaneous tag/value pairs
#   mbRenameIf[] - array of renamable interfaces
#   mbHideIf[] - array of hidable interfaces
#   res - 0 if the script doesn't execute to completion (error), 1 if it
#         executed successfully, 2 if executed to completion, but there
#         was an error that was logged to syslog
#____________________________________________________________________
{
    #See SDD for tag syntax
    #if it is not a comment line
    if (match ($1, /^[ \t]*#/) == 0)
    {
        #if it is not a "match." type tag
        if (match ($1, /^[ \t]*match\./) == 0)
        {
            #Normalize the fields.  We expect a tag containing no '=',
            # followed by any value string.  Since the value could contain 
            # an '=' and we have to check that there are only 2 fields
            # found in this record, and if there are more, concatenate
            # $2 with the additional ones (inserting a '=')
            if ( (length ($1) != 0) && (length ($2) == 0)) 
            {
                system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Tag missing value (" $1 "); ignoring it\"")
            }
            else
            {
                field2 = $2
                if ( NF > 2 )
                {
                    for (i=3; i<=NF; ++i)
                    {
                       field2 = field2 "=" $i
                    }
                }

                #convert " to \"
                pattern = "\""
                if (match (field2, pattern))
                {
                    str1 = substr( field2, 1, RSTART-1 )
                    str2 = substr( field2, RSTART+1 )
                    field2 = str1 "\\\""
                    while (match (str2, pattern))
                    {
                        str1 = substr( str2, 1, RSTART-1 )
                        str2 = substr( str2, RSTART+1 )
                        field2 = field2 str1 "\\\""
                    }
                    field2 = field2 str2 
                }

                #strip off the trailing newline
                if (match (field2, /\n$/)) 
                {
                    len = length ( field2 ) - 1
                    field2 = substr( field2, 1, len )
                }

                #Now parse the tags
                if ( match ($1, /^[ \t]*mobo.model/) )
                {
                    CheckForDupes( $1, field2 )
                    #it is a model string tag
                    if( mbModel == "none specified" )
                    {
                        mbModel = field2
                    }
                    else
                    {
                        mbModel = mbModel " or " field2
                    }
                }
                else if ( match ($1, /^[ \t]*mobo.ports/) )
                {
                    #It's a ports tag
                    if ( match ($1, /^[ \t]*mobo.ports.count/) )
                    {
                        CheckForDupes( $1, field2 )
                        #must be a number
                        if ( match (field2, /^[ \t]*[0-9]+[ \t]*$/) ) 
                        {
                            mbNumPorts = field2
                            readNumPorts = 1
                        }
                        else
                        {
                            system ("logger -p local0.error \"" logPrefix "(" argFileName ") Invalid tag: " $1 "=" field2 "\"")
                            res = 2
                        }
                    }
                    else if ( offset = match ($1, /^[ \t]*mobo.ports.port/) )
                    {
                        CheckForDupes( $1, field2 )
                        portsCounted++
                        #it's a port<N> spec
                        offset = offset + RLENGTH
                        len = length ( $1 ) - offset + 1
                        portnum = substr ( $1, offset, len )
                        mbPort[portnum] = field2
                    }
                    else if ( offset = match ($1, /^[ \t]*mobo.ports.names/) )
                    {
                        CheckForDupes( $1, field2 )
                        #it's a CSL of port names
                        mbNumPortNames = split( field2, names, "," )
                        for (i=0; i<mbNumPortNames; ++i) 
                        {
                            if( match( names[i+1], /^[ \t]*[A-z]+[ \t]*$/ ) )
                            {
                                mbPortNames[i] = substr( names[i+1], RSTART, RLENGTH )
                            }
                            else
                            {
                                system ("logger -p local0.error \"" logPrefix "(" argFileName ") Invalid tag: " $1 "=" field2 "; " names[i+1] " must be alphabetic characters\"")
                                res = 2
                            }
                        }
                    }
                    else
                    {
                        system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Unrecognized tag found (" $1 ")\"")
                    }
                }
                else if ( match ($1, /^[ \t]*mobo.pciexp/) )
                {
                    #it's a pci expansion slot tag
                    if ( match ($1, /^[ \t]*mobo.pciexp.count/) )
                    {
                        CheckForDupes( $1, field2 )
                        #must be a number
                        if ( match (field2, /^[ \t]*[0-9]+[ \t]*$/) ) 
                        {
                            mbNumSlots = field2
                            readNumSlots = 1
                        }
                        else
                        {
                            system ("logger -p local0.error \"" logPrefix "(" argFileName ") Invalid tag: " $1 "=" field2 "\"")
                            res = 2
                        }
                    }
                    else if ( offset = match ($1, /^[ \t]*mobo.pciexp.slot/) )
                    {
                        #it's a slot<N> tag
                        CheckForDupes( $1, field2 )
                        slotsCounted++
                        offset = offset + RLENGTH
                        len = length ( $1 ) - offset + 1
                        slotnum = substr ( $1, offset, len )
                        mbSlot[slotnum] = field2
                    }
                    else if ( offset = match ($1, /^[ \t]*mobo.pciexp.reversed/) )
                    {
                        CheckForDupes( $1, field2 )
                        #it's a reversed<N> tag
                        offset = offset + RLENGTH
                        len = length ( $1 ) - offset + 1
                        slotnum = substr ( $1, offset, len )
                        if (match (field2, /^[ \t]*0[ \t]*$/) )
                        {
                            mbSlotRev[slotnum] = 0
                        }
                        else if (match (field2, /^[ \t]*1[ \t]*$/) ) 
                        {
                            mbSlotRev[slotnum] = 1
                        }
                        else
                        {
                            system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Invalid tag (treating as 1): " $1 "=" field2 "\"")
                            mbSlotRev[slotnum] = 1
                        }
                    }
                    else
                    {
                        system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Unrecognized tag found (" $1 ")\"")
                    }
                }
                else if ( match ($1, /^[ \t]*mobo.bios/) )
                {
                    if ( offset = match ($1, /^[ \t]*mobo.bios.nmismisupport.date/) )
                    {
                        CheckForDupes( $1, field2 )
                        mbNmiSmiDate = field2
                    }
                    else if ( offset = match ($1, /^[ \t]*mobo.bios.nmismisupport.rev/) )
                    {
                        CheckForDupes( $1, field2 )
                        mbNmiSmiRev = field2
                    }
                    else if ( offset = match ($1, /^[ \t]*mobo.bios.nmismisupport.getdate/) )
                    {
                        CheckForDupes( $1, field2 )
                        mbNmiSmiGetDate = field2
                    }
                    else if ( offset = match ($1, /^[ \t]*mobo.bios.nmismisupport.getrev/) )
                    {
                        CheckForDupes( $1, field2 )
                        mbNmiSmiGetRev = field2
                    }
                    else
                    {
                        system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Unrecognized tag found (" $1 ")\"")
                    }
                }
                else if ( match ($1, /^[ \t]*misc[ \t]*$/) )
                {
                    mbMisc[mbNumMisc] = field2
                    ++mbNumMisc
                }
                else if ( match ($1, /^[ \t]*renameif[ \t]*$/) )
                {
                    mbRenameIf[mbNumRenameIf] = field2
                    ++mbNumRenameIf
                }
                else if ( match ($1, /^[ \t]*hideif[ \t]*$/) )
                {
                    mbHideIf[mbNumHideIf] = field2
                    ++mbNumHideIf
                }
                else if ( match ($1, /^[ \t]*include[ \t]*$/) )
                {
		    #ignore includes
                }
                else if ( match ($1, /^[ \t]*$/ ) ) 
                {
                    #ignore blank lines
                }
                else 
                {
                    #anything else is unrecognized
                    system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Unrecognized tag found (" $1 ")\"")
                }
            }
        }
    }
}


