#!/bin/bash
#set -x
#
# Must be called with either stop or start.
#

LogMessage()
{
    msg=$(echo "mobomap: ";echo $2)
    logger -p user.$1 $msg
}
LogError()
{
    LogMessage 4 "$1"
    if [ $verbose = 1 ]; then echo "syslog: Error: $1"; fi
}
LogWarning()
{
    LogMessage 5 "$1"
    if [ $verbose = 1 ]; then echo "syslog: Warning: $1"; fi
}
LogNotice()
{
    LogMessage 6 "$1"
    if [ $verbose = 1 ]; then echo "syslog: Notice: $1"; fi
}
LogInfo()
{
    LogMessage 7 "$1"
    if [ $verbose = 1 ]; then echo "syslog: Info: $1"; fi
}
Bail()
{
    LogError "Motherboard map file was not generated"
    exit 1
}

Usage()
{
    echo "Usage: $0 [-v] [-i] [-r] [-p path]"
    echo "Usage: $0 [-h]"
    echo ' v - verbose'
    echo ' i - ignore existing map'
    echo ' r - use testpath (if given) for rc.conf.local'
    echo ' p - path to use as the root (for testing)'
    echo ' h - help; prints this message and exits'
    echo ' s - stop; used to undo renaming of the interfaces'
    echo ''

    exit 2
}

DoQuotes()
{
    #if $1 contains whitespace, wrap it with quotes
    echo $(echo $1 | awk -F"\n" '{
            if (match ($1, /[ \t]/))
            {
                print "\"" $1 "\""
            }
            else
            {
                print $1
            }
        }')
}

myAppName=`basename $0`

if [ -z "$install_base" ]
then
    install_base=$(dirname $0)
    BASE=$install_base/etc/mobomap
    MAPBASE=$install_base/etc
    MOBOMAPDIR=$install_base/run
fi

if [ -d /etc/mobomap ]
then
    BASE=/etc/mobomap
    install_base=
    MAPBASE=/etc
    MOBOMAPDIR=/run
fi

if [ -d /usr/local/sandvine/etc ]
then
    BASE=/usr/local/sandvine/etc/mobomap
    install_base=/usr/local/sandvine
    MAPBASE=/usr/local/sandvine/etc
    MOBOMAPDIR=/usr/local/sandvine/run
fi

# Platform mobomap located in /etc/sv/mobomap
if [ `dirname $0` = "/etc/sv/mobomap" ] ; then
    install_base=$(dirname $0)
    BASE=$install_base
    MAPBASE=$install_base/etc
    MOBOMAPDIR=$install_base
fi

let verbose=0
let checkPrevious=1
let useTestPath=0
let useTestPathForRc=0
stopflag=0
ifRenameFile=

while getopts "hin:p:rsv" opt; do
    case "$opt" in
    h)
	Usage
	;;
    i)
	checkPrevious=0
	;;
    p)
	testPath="$OPTARG"
	useTestPath=1
	;;
    n)
	ifRenameFile="$OPTARG"
	;;
    r)
	useTestPathForRc=1
	;;
    s)
	stopflag=1
	;;
    v)
	verbose=1
	;;
    *)
	Usage
	;;
    esac
done
shift $(($OPTIND - 1))


if [ $verbose = 1 ]; then
    echo "verbose=$verbose"
    echo "checkPrevious=$checkPrevious"
    echo "useTestPathForRc=$useTestPathForRc"
    if [ $useTestPath = 1 ]; then
        echo "testPath=$testPath"
    fi
fi

rcConfLocalPath=/etc

if [ $useTestPath = 1 ]; then
    MAPBASE=${testPath}
    BASE=${testPath}/mobomap
    if [ $useTestPathForRc = 1 ]; then
        rcConfLocalPath=${testPath}
    fi
fi

outFile=${MOBOMAPDIR}/motherboard.map
moboCfgDir=${BASE}/config/motherboard
moboIncludeDir=${BASE}/config/include
cardCfgDir=${BASE}/config/card
diskMapDir=${BASE}/config/diskmap

if [ $verbose = 1 ]; then
    echo "Motherboard map output file is: $outFile"
    echo "Motherboard cfg file dir is: $moboCfgDir"
    echo "Card cfg file dir is: $cardCfgDir"
fi


#Keep \n when we store results of scripts and executables so we can properly
# awk them
OLDIFS="$IFS"
IFS=" "
sysinfoRes=$(/usr/local/bin/sysinfo | awk -f $BASE/fixNewlines.awk)
sysctlRes14k=$(sysctl -a | grep 14k | awk -f $BASE/fixNewlines.awk)
dmesgRes=$(dmesg | awk -f $BASE/fixNewlines.awk)
pciconfResOrig=$(pciconf -l -v)
pciconfRes=$(pciconf -l -v | awk -f $BASE/fixNewlines.awk)
pciconfResNoDetailsOrig=$(pciconf -l)
pciconfResNoDetails=$(pciconf -l | awk -f $BASE/fixNewlines.awk)
pciconfRemaining=$pciconfResNoDetailsOrig
matchFile=0
IFS="$OLDIFS"


unset mbCfgFileMatch
#For each file in the motherboard config file directory, try to match the tags
# in the file with the results of the system utilities executed above
for file in $(ls $moboCfgDir) ; do
    if [ $verbose = 1 ]; then echo "Working file $file"; fi

    #Run the match script - it will set mbCfgFileMatch to 1 if it matched
    # the file, 2 if it did not, and 0 if there was some error in script
    mbCfgIncludeFile=""
    eval $(awk -f $BASE/matchMoboToCfgFile.awk "argPciconfRes=$pciconfRes" "argSysinfoRes=$sysinfoRes" "argDmesgRes=$dmesgRes" "argFileName=$file" $moboCfgDir/$file)
    if [ "$mbCfgFileMatch" = "1" ]; then
        if [ $verbose = 1 ]; then
            echo " "
            echo "Matched this mobo to $file"
        fi
        LogInfo "Matched the motherboard to config file $file"
        matchFile=$file
        includeFiles=$mbCfgIncludeFile
        break
    elif [ "$mbCfgFileMatch" = "2" ]; then
        if [ $verbose = 1 ]; then echo "failed tag: $failedTag"; fi
    elif [ "$mbCfgFileMatch" = "0" ]; then
        if [ $verbose = 1 ]; then echo "Error"; fi
        LogError "Unknown error encountered running match script on $file"
    fi
done

if [ "$matchFile" = "0" ]; then
    LogError "No matching config file was found for this motherboard"
    Bail
fi

if [ $verbose = 1 ]; then
    echo " "
    echo "-------------------------------------------------------------------------"
    echo "                     Parsing Motherboard Config File"
    echo "-------------------------------------------------------------------------"
    echo "Parsing $matchFile for config"
fi
includeStr=
for file in $includeFiles; do
   includeStr=$includeStr" "$moboIncludeDir/$file
done
#Extract all the configuration information from the config file.  See the awk
# scripts comments for a list of the shell variables that are set by the script.
eval $(awk -f $BASE/extractMoboCfgInfo.awk "argPciconfRes=$pciconfRes" "argSysinfoRes=$sysinfoRes" "argFileName=$matchFile" $moboCfgDir/$matchFile $includeStr)

#If there was an error parsing the motherboard config, we're done
if [ "$res" = "1" ]; then
    if [ $verbose = 1 ]; then
        echo "Successfully parsed motherboard config file"
    fi
elif [ "$res" = "2" ]; then
    #message has already been logged
    Bail
else
    if [ $verbose = 1 ]; then
        echo "Unknown error occurred while trying to extract motherboard config info from $matchFile"
    fi
    LogError "Unknown script error while trying to parse motherboard config from $matchFile"
    Bail
fi

# check for any expansion cards
for file in $(ls $cardCfgDir | grep exp) ; do
   if [ $verbose = 1 ]; then echo "Working card file $file"; fi

   #Run the match script - it will set crdCfgFileMatch to 1 if it matched
   # the file, 2 if it did not, and 0 if there was some error in script
   eval $(awk -f $BASE/matchExpansionToCfgFile.awk "argSysctl=$sysctlRes14k" "argFileName=$file" $cardCfgDir/$file)

   if [ "$expansionCfgFileMatch" = "1" ]; then
      if [ $verbose = 1 ]; then
         echo " "
         echo "Matched this card to $file"
      fi
      LogInfo "Using $file for config of expansion card"
      eval $(awk -f $BASE/extractExpansionCfgInfo.awk "argFileName=$file" $cardCfgDir/$file)
      #If there was an error parsing the card config, we're done
      if [ "$res" = "1" ]; then
         if [ $verbose = 1 ]; then
            echo "Successfully parsed card config file"
         fi
      elif [ "$res" = "2" ]; then
         #message has already been logged
         Bail
      else
         if [ $verbose = 1 ]; then
            echo "Unknown error occurred while trying to extract card config info from $file"
         fi
         LogError "Unknown script error while trying to parse card config from $file"
         Bail
      fi
      break
   elif [ "$expansionCfgFileMatch" = "2" ]; then
      if [ $verbose = 1 ]; then echo "failed tag: $failedTag"; fi
   elif [ "$expansionCfgFileMatch" = "0" ]; then
      if [ $verbose = 1 ]; then echo "Error"; fi
      LogError "Unknown error while trying to match $file"
   fi
done



if [ $verbose = 1 ]; then
    echo "Motherboard is $mbModel"
    echo "There are $mbNumPorts ports"
    for i in "${mbPort[@]}"; do
        echo "   port is $i"
    done

    echo "There are $mbNumSlots pci slots"
    for i in "${mbSlot[@]}"; do
        echo "   slot is select $i"
    done

    for i in "${mbSlotRev[@]}"; do
        echo "   slot rev is $i"
    done

    echo "mobo port names are:"
    for i in "${mbPortNames[@]}"; do
        echo "   $i"
    done

    echo "Miscellaneous Tags"
    for i in "${mbMisc[@]}"; do
        echo "   $i"
    done

    echo "Expansion Miscellaneous Tags"
    for i in "${expMisc[@]}"; do
        echo "   $i"
    done

    echo "RenameIF Tags"
    for i in "${mbRenameIf[@]}"; do
        echo "   $i"
    done

    echo "Expansion RenameIF Tags"
    for i in "${expRenameIf[@]}"; do
        echo "   $i"
    done

    echo "HideIF Tags"
    for i in "${mbHideIf[@]}"; do
        echo "   $i"
    done

    echo "Expansion hideIF Tags"
    for i in "${expHideIf[@]}"; do
        echo "   $i"
    done

    echo "NMI/SMI Min Date is $mbNmiSmiDate"
    echo "NMI/SMI Min Rev is $mbNmiSmiRev"
    echo "NMI/SMI Date Get script is $mbNmiSmiGetDate"
    echo "NMI/SMI Rev Get script is $mbNmiSmiGetRev"
fi

#set -x
if [ "$mbNmiSmiDate" != "none specified" ]; then
    if [ "$mbNmiSmiGetDate" != "none specified" ]; then
        date=$(sh -c "$mbNmiSmiGetDate")
        if [ $verbose = 1 ]; then
            echo "date=$date, minDate=$mbNmiSmiDate"
        fi
        if [ "$date" \> "$mbNmiSmiDate" ]; then
            sysctl hw.watchdog.tco_watchdog_enable_tco_smi_int=1 >/dev/null 2>&1
        fi
    else
        LogError "Min NMI/SMI date given ($mbNmiSmiDate) but no way to get the date was given"
        Bail
    fi
elif [ "$mbNmiSmiRev" != "none specified" ]; then
    if [ "$mbNmiSmiGetRev" != "none specified" ]; then
        rev=$(sh -c "$mbNmiSmiGetRev")
        if [ $verbose = 1 ]; then
            echo "rev=$rev, minRev=$mbNmiSmiRev"
        fi
        if [ "$rev" \> "$mbNmiSmiRev" ]; then
            sysctl hw.watchdog.tco_watchdog_enable_tco_smi_int=1 >/dev/null 2>&1
        fi
    else
        LogError "Min NMI/SMI rev given ($mbNmiSmiRev) but no way to get the rev was given"
        Bail
    fi
fi


#portCount starts as the number of motherboard ports (as specified by the
#config file.  As ports on cards are discovered, it is incremented.
let portCount=$mbNumPorts
m_numPortsBySlot[0]=$mbNumPorts

#Our current slot number
let slotNum=0

#Slots can be listed and wired in any order.  PCI selects specified for slots
# assume that there are no bridges on the cards in the slot.  In the event that
# a bridge is encountered, "later" slots (ones with a larger bus number - we can
# ignore the device number here as peer devices will be unaffected by bridges
# on other peers) must have their bus number incremented by one for each bridge
# encountered in an earlier PCI slot.  So we need to generate an ordered index
# of slot numbers.
IFS=" "
let i=0
for slot in "${mbSlot[@]}"; do
    let i=$i+1
    mbSlotArg=$(echo "$mbSlotArg";echo "$i $slot")
done
eval $(echo $mbSlotArg | sort -n | awk '
    BEGIN{
        count = 0
    }
    END{
        for (i=0; i<count; ++i)
        {
            printf( "mbSlotOrderedIdx[%s]=\"%s\"\n", i, mbSlotOrderedIdx[i] )
        }
    }{
        if (NF > 1)
        {
            mbSlotOrderedIdx[count] = $1
            ++count
        }
    }')
IFS="$OLDIFS"

#Now we do each slot as ordered by their bus numbers.  Note, because multiple
# select lines can be specified for each slot, we have made an assumption that
# the select lines will be different devices on the same bus.  If this rule
# is violated, the algorithm might need to modified to include the other selects
# in the sort.
for slotNum in "${mbSlotOrderedIdx[@]}"; do
    #slot is the partial pci select (pci<bus>:<device>)
    slot=${mbSlot[$slotNum]}
    if [ $verbose = 1 ]; then
        echo " "
        echo "-------------------------------------------------------------------------"
        echo "                           Analyzinging Slot $slotNum"
        echo "-------------------------------------------------------------------------"
    fi

    #unset the variables we'll be using in the loop
    #cardFile is the name of the config file
    unset cardFile
    #cardModel is the model name of the card
    unset crdModel
    #slotIdx is an array of indices used to iterate over each of the following
    # arrays
    unset slotIdx
    #slotDevs is an array devices in this slot
    unset slotDevs
    #slotBridge is a array of flags that indicate whether the device is a bridge
    unset slotBridge
    #slotBusses is an array of bus IDs as in pci<bus>:0:0
    unset slotBusses
    #slotIdx is an array of device IDs as in pci0:<device>:0
    unset slotDevices
    #slotIdx is an array of functions as in pci0:0:<function>
    unset slotFns
    #mbPciSel is an array of parsed pci selectors
    unset mbPciSel

    #the number of devices in this slot
    let slotNumDevs=0
    if [ $verbose = 1 ]; then echo "working slot $slot (as specified by cfg file, before accounting for bridges on earlier slots)"; fi

    #PCI selector can be a CSL, so we have to split them up and stuff them
    # in mbPciSel
    eval $(echo $slot | awk -F"[ \t]*,[ \t]*" '{for (i=1; i<=NF; ++i) {printf ("mbPciSel[%s]=\"%s\"\n",i,$i)} }')

    #for each select line on the slot
    for sel in "${mbPciSel[@]}"; do
        #unset the number of devices, index array, device array, bridge flags
        # array, bus numbers array, device numbers array, and functions array
        # for this PCI selector
        unset mbNumDevs
        unset mbIdx
        unset mbDevs
        unset mbBridge
        unset mbBusses
        unset mbDevices
        unset mbFns

        #Extract this selector's bus number and assign to bus
        eval $(echo "$sel" | awk -F"\n" '{
                match ($1, /[0-9]+/)
                bus = int( substr( $1, RSTART, RLENGTH ) )
                printf ("bus=\"%s\"\n", bus )
            }')

        #Count the number of bridges that appear in PCI slots whose bus number
        # is smaller than ours.  Each one that we find will cause the bus
        # number specified in the config file to be incremented by one.
        let inc=0
        for ((i=0; $i<$bus; i=$i+1)); do
            let inc=${mbNumCardBridges[${i}]}+${inc}
        done

        #Increment bus number of our selector by inc and assign to modSel
        eval $(echo "$sel" | awk -F"\n" '{
                match ($1, /[0-9]+/)
                bus = int( substr( $1, RSTART, RLENGTH ) )
                bus = bus + inc
                printf ("modSel=\"pci%s%s\"\n", bus, substr($1, RSTART+RLENGTH) )
            }' "inc=$inc")


        #Get all the devices that are on the modSel selector
        eval $(echo "$pciconfResOrig" | awk -f $BASE/getDevicesOnPciSelect.awk "argPciSel=$modSel" "argStartCount=0" "argBusOnly=0")

        #For each device...
        for dev in "${mbIdx[@]}"; do
            #Append the devices we found to the array of devices in the slot
            # (remeber one slot can have multiple selects)
            slotIdx[$slotNumDevs]=${mbIdx[$dev]}
            slotDevs[$slotNumDevs]=${mbDevs[$dev]}
            slotBridge[$slotNumDevs]=${mbBridge[$dev]}
            slotBusses[$slotNumDevs]=${mbBusses[$dev]}
            slotDevices[$slotNumDevs]=${mbDevices[$dev]}
            slotFns[$slotNumDevs]=${mbFns[$dev]}
            #increment our count of devices in the slot
            let slotNumDevs=$slotNumDevs+1

            #If it's a bridge device we have to find the devices on the far side
            # and add them to the list of devices in the slot
            if [ ${mbBridge[$dev]} = 1 ]; then
                #Can simplify these three statements, but there are issues re:
                # the val="" case
                let x=mbNumCardBridges[${mbBusses[$dev]}]
                let x=$x+1
                mbNumCardBridges[${mbBusses[$dev]}]=$x

                #Get the secondary bus number and the corresponding pciSel
                mbSecBusNum=`pciconf -r -b $modSel:${mbFns[$dev]} 0x19`
                # Pciconf output is in hex, but may or may not have the 0x
                # prefix depending on the specific version.  Remove it if
                # it's there, then add it back and use printf to convert to
                # decimal.
                mbSecBusNum=0x${mbSecBusNum#0x}
                pciSel=`printf "pci%d" $mbSecBusNum`

                #back up these arrays so the data we're using isn't clobbered
                # below
                mbNumDevsBak=$mbNumDevs
                mbIdxBak=( ${mbIdx[@]} )
                mbDevsBak=( ${mbDevs[@]} )
                mbBridgeBak=( ${mbBridge[@]} )
                mbBussesBak=( ${mbBusses[@]} )
                mbDevicesBak=( ${mbDevices[@]} )
                mbFnsBak=( ${mbFns[@]} )
                #Clear those variables we've been using to get at the devices
                # in a selector
                unset mbNumDevs
                unset mbIdx
                unset mbDevs
                unset mbBridge
                unset mbBusses
                unset mbDevices
                unset mbFns

                #Get all the devices on the new selector (note we have to set
                # argBusOnly=1)
                eval $(echo "$pciconfResOrig" | awk -f $BASE/getDevicesOnPciSelect.awk "argPciSel=$pciSel" "argStartCount=0" "argBusOnly=1" "mbNumCardBridges=$mbNumCardBridges")

                #For each device on the far side of the bus
                for m in "${mbIdx[@]}"; do
                    #Add to our list of devices in the slot
                    slotIdx[$slotNumDevs]=$slotNumDevs
                    slotDevs[$slotNumDevs]=${mbDevs[$m]}
                    slotBridge[$slotNumDevs]=${mbBridge[$m]}
                    slotBusses[$slotNumDevs]=${mbBusses[$m]}
                    slotDevices[$slotNumDevs]=${mbDevices[$m]}
                    slotFns[$slotNumDevs]=${mbFns[$m]}

                    #increment our count of devices in the slot
                    let slotNumDevs=$slotNumDevs+1
                done

                #reset the backed-up vars
                unset mbNumDevs
                unset mbIdx
                unset mbDevs
                unset mbBridge
                unset mbBusses
                unset mbDevices
                unset mbFns
                mbNumDevs=$mbNumDevsBak
                mbIdx=( ${mbIdxBak[@]} )
                mbDevs=( ${mbDevsBak[@]} )
                mbBridge=( ${mbBridgeBak[@]} )
                mbBusses=( ${mbBussesBak[@]} )
                mbDevices=( ${mbDevicesBak[@]} )
                mbFns=( ${mbFnsBak[@]} )
                unset mbNumDevsBak
                unset mbIdxBak
                unset mbDevsBak
                unset mbBridgeBak
                unset mbBussesBak
                unset mbDevicesBak
                unset mbFnsBak
            fi
        done

        if [ $verbose = 1 ]; then
            echo "Found $slotNumDevs total devs at pci select $modSel"
            for dev in "${slotIdx[@]}"; do
                echo "   $dev) pci${slotBusses[$dev]}:${slotDevices[$dev]}:${slotFns[$dev]}=${slotDevs[$dev]}"
            done
        fi
    done

    #These are used to pass in our arrays and variables into scripts...they're
    # simply newline separated lists of values.  Because of the way they're
    # constructed, the first element is ""...this is useful since most of them
    # are supposed to be counted starting at 1.
    unset argNumDevs
    unset argIdx
    unset argDevs
    unset argBridge
    unset argBusses
    unset argDevices
    unset argFns
    unset argPciconf
    argNumDevs="${slotNumDevs}"

    #Remember, clearing the IFS flag ensures that \n doesn't get converted to
    # a space.  We reset it once the variables are set.
    IFS=" "
    for dev in "${slotIdx[@]}"; do
        #slotPciconf is the pciconf results using the -r (read) option on the
        # pci selector.  We read the entire 256 byte config space and pass
        # it in so that the awk script can pull out the bits it might need to
        # match the card.
        slotPciconf[$dev]=$(pciconf -r pci${slotBusses[$dev]}:${slotDevices[$dev]}:${slotFns[$dev]} 0:63 | awk -F"\n" '{if (match ($0, " $")) {printf("%s",$0)} else {printf("%s ",$0)}}' )
        argIdx=$(echo "$argIdx";echo "${slotIdx[$dev]}")
        argDevs=$(echo "$argDevs";echo "${slotDevs[$dev]}")
        argBridge=$(echo "$argBridge";echo "${slotBridge[$dev]}")
        argBusses=$(echo "$argBusses";echo "${slotBusses[$dev]}")
        argDevices=$(echo "$argDevices";echo "${slotDevices[$dev]}")
        argFns=$(echo "$argFns";echo "${slotFns[$dev]}")
        argPciconf=$(echo "$argPciconf";echo "${slotPciconf[$dev]}")
    done
    IFS="$OLDIFS"

    if [ $verbose = 1 ]; then
        echo " "
        echo "-------------------------------------------------------------------------"
        echo "                 Matching Config File for Card in Slot $slotNum"
        echo "-------------------------------------------------------------------------"
    fi

    argIdx=$(echo "$argIdx" | awk -f $BASE/fixNewlines.awk)
    argDevs=$(echo "$argDevs" | awk -f $BASE/fixNewlines.awk)
    argBridge=$(echo "$argBridge" | awk -f $BASE/fixNewlines.awk)
    argBusses=$(echo "$argBusses" | awk -f $BASE/fixNewlines.awk)
    argDevices=$(echo "$argDevices" | awk -f $BASE/fixNewlines.awk)
    argFns=$(echo "$argFns" | awk -f $BASE/fixNewlines.awk)
    argPciconf=$(echo "$argPciconf" | awk -f $BASE/fixNewlines.awk)

    #For each configuration file in the card directory
    for file in $(ls $cardCfgDir) ; do
        if [[ $file = *exp* ]]; then
           continue
        fi

        if [ $verbose = 1 ]; then echo "Working card file $file"; fi

        #Run the match script - it will set crdCfgFileMatch to 1 if it matched
        # the file, 2 if it did not, and 0 if there was some error in script
        eval $(awk -f $BASE/matchCardToCfgFile.awk "argNumDevs=$argNumDevs" "argIdx=$argIdx" "argDevs=$argDevs" "argBridge=$argBridge" "argBusses=$argBusses" "argDevices=$argDevices" "argFns=$argFns" "argPciconf=$argPciconf" "argFileName=$file" $cardCfgDir/$file)

        if [ "$crdCfgFileMatch" = "1" ]; then
            if [ $verbose = 1 ]; then
                echo " "
                echo "Matched this card to $file"
            fi
            LogInfo "Using $file for config of card in slot $slotNum"
            cardFile=$file
            break
        elif [ "$crdCfgFileMatch" = "2" ]; then
            if [ $verbose = 1 ]; then echo "failed tag: $failedTag"; fi
        elif [ "$crdCfgFileMatch" = "0" ]; then
            if [ $verbose = 1 ]; then echo "Error"; fi
            LogError "Unknown error while trying to match $file to card in slot $slotNum"
        fi
    done

    if [ $verbose = 1 ]; then
        echo " "
        echo "-------------------------------------------------------------------------"
        echo "                 Parsing Config File for Card in Slot $slotNum"
        echo "-------------------------------------------------------------------------"
    fi

    #If we matched the card to the file
    if [ "$crdCfgFileMatch" = "1" ]; then
        if [ $verbose = 1 ]; then echo "Parsing $cardFile for config"; fi

        #Extract the configuration info from the config file and assign it
        # to shell variables.  See script comments for variable names
        eval $(awk -f $BASE/extractCardCfgInfo.awk "argReversed=${mbSlotRev[slotNum]}" "argFileName=$cardFile" $cardCfgDir/$cardFile)
        if [ $verbose = 1 ]; then
            echo "crdNumPorts = $crdNumPorts"
            echo "crdModel = $crdModel"
        fi
        m_numPortsBySlot[$slotNum]=$crdNumPorts

        #If there was an error parsing the card config, we're done
        if [ "$res" = "1" ]; then
            if [ $verbose = 1 ]; then
                echo "Successfully parsed card config file"
            fi
        elif [ "$res" = "2" ]; then
            #message has already been logged
            Bail
        else
            if [ $verbose = 1 ]; then
                echo "Unknown error occurred while trying to extract card config info from $cardFile"
            fi
            LogError "Unknown script error while trying to parse card config from $cardFile"
            Bail
        fi

        #For each physical port on the card (elements of this array are the
        # offset of the devices in this slot where we would find the device
        # matching this physical port)
        for p in "${crdPort[@]}"; do
            isNonNull=0
            eval $(echo ${slotDevs[p]} | awk -F"\n" '{if (match ($1, /[0-z]+/)) {printf("isNonNull=1\n")}}')
            if [ $isNonNull = 1 ]; then
                #increment our total port count (includes mobo ports)...
                let portCount=$portCount+1
                #...and add it to the list of ports (remember slotDevs is the
                # list of device names in this slot)
                if [ "x$crdNameMapping" == "x" ]; then
                    mbPort[$portCount]=${slotDevs[p]}
                else
                    mbPort[$portCount]="${crdNameMapping}$(echo ${slotDevs[p]} | awk '{if (match( $0, /[0-9]+/ )) {print substr($0, RSTART, RLENGTH)}}')"
                fi
            fi
        done

        #Store the config file for this slot in the outputSlotCfgFile array
        outputSlotCfgFile[$slotNum]=$cardFile
        #Store away the card model.
        outputSlotModel[$slotNum]=$crdModel
        #Extract the pciconf info that pertains only to this slot
        unset argSlotDevs
        IFS=" "
        let count=1
        for d in "${slotDevs[@]}"; do
            let count=$count+1
            argSlotDevs=$(echo $argSlotDevs; echo $d)
        done
        argSlotDevs=$(echo $argSlotDevs | awk -f $BASE/fixNewlines.awk)
        eval $(echo $pciconfRemaining | awk -f $BASE/extractPciconfForDevices.awk "argSlotDevs=$argSlotDevs" "argNumDevs=$count")

        #Now let's generate the signature using the MD5 hash utility.
        outputSlotSig[$slotNum]=$(md5 -q -s "${pciconfForThisSlot}")
        IFS="$OLDIFS"
    elif [ "$crdCfgFileMatch" = "0" ]; then
        if [ $verbose = 1 ]; then
            echo "ERROR matching config files for card in slot $slotNum"
        fi
        LogError "Unknown error occurred while matching config files for slot $slotNum"
    else
        if [ $slotNumDevs == 0 ]; then
            if [ $verbose = 1 ]; then
                echo "No card in slot $slotNum"
            fi
            LogInfo "No card was found in slot $slotNum"
        else
            if [ $verbose = 1 ]; then
                echo "No matching config file was found for card in slot $slotNum."
                echo "It's either not a NIC, or not one we recognize."
            fi
            LogError "No matching config file was found for card in slot $slotNum."
        fi
    fi
done

#Whatever's left in the pciconfRemaining after looking at all the slots must
# be part of the mobo and will be used to generate the sig
IFS=" "
mobosig=$(md5 -q -s "${pciconfRemaining}")
IFS="$OLDIFS"

#Stuff our port list (all physical ports including slots and mobo) and the
# device names into arg vars to pass into the assignRemainingPorts script
IFS=" "
for val in "${mbPort[@]}"; do
    argMbPort=$(echo "$argMbPort";echo "${val}")
done
let argNumNames=0
for val in "${mbPortNames[@]}"; do
    argPortNames=$(echo "$argPortNames";echo "${val}")
    let argNumNames=$argNumNames+1
done
let argNumExpPorts=portCount-mbNumPorts

#Assign remaining ports takes the array of ports and the device name list and
# fills in the missing info.  At this point the port lists has the correct
# device names filled into mbPort[] for all of the expansion slot ports, but
# only the device offsets for the motherboard ports. This is because until we've
# accounted for all of the slot devices, we don't know which ones remain (these
# are the mobo devices).
argMbPort=$(echo $argMbPort | awk -f $BASE/fixNewlines.awk)
argPortNames=$(echo $argPortNames | awk -f $BASE/fixNewlines.awk)
eval $(echo $pciconfResNoDetailsOrig | awk -f $BASE/assignRemainingPorts.awk "argNumMoboPorts=$mbNumPorts" "argNumExpPorts=$argNumExpPorts" "argMbPort=$argMbPort" "argNumNames=$argNumNames" "argPortNames=$argPortNames")
IFS="$OLDIFS"

if [ $verbose = 1 ]; then
    echo " "
    echo "-------------------------------------------------------------------------"
    echo "                     Generating Motherboard Map File"
    echo "-------------------------------------------------------------------------"
fi
if [ $verbose = 1 ]; then
    let x=0
    echo ""
    for ports in "${mbPort[@]}"; do
       let x=$x+1
       echo "Port$x-$ports"
    done
fi

if [ $stopflag = 1 ]; then
    for i in "${mbRenameIf[@]}"; do
        fromName=$(echo $i | sed 's/=.*//')
        toName=$(echo $i | sed 's/.*=//')
        eval ifconfig $toName name $fromName >/dev/null 2>&1
    done

    for i in "${expRenameIf[@]}"; do
         fromName=$(echo $i | sed 's/=.*//')
         toName=$(echo $i | sed 's/.*=//')
         eval ifconfig $toName name $fromName >/dev/null 2>&1
    done

    for i in "${mbHideIf[@]}"; do
        eval ifconfig $i -hidden >/dev/null 2>&1
    done

    for i in "${expHideIf[@]}"; do
         eval ifconfig $i -hidden >/dev/null 2>&1
    done

    exit 0
fi

#Generate the output
matchFile=$(DoQuotes "$matchFile")
mbModel=$(DoQuotes "$mbModel")
echo "#Sandvine Motherboard map" > $outFile
echo "#Automatically generated" >> $outFile
echo "#" >> $outFile

#Make the ascii pic
echo "#Back of box looks like:" >> $outFile
let i=1
let portsDone=0
line2=""
line1=""
let first=1
for numPorts in "${m_numPortsBySlot[@]}"; do
    let todo=$i+$numPorts
    for ((j=$i; $j<$todo; j=$j+1)); do
        if [ $first = "1" ]; then
            line2="${line2}[${mbPort[$i]}]"
        else
            line1="[${mbPort[$i]}]$line1"
        fi
        let i=$i+1
    done
    if [ $first = "1" ]; then
        let first=0
    else
        line1="_______${line1}"
    fi
done
lenOfFirstBit=$(echo $line1 | awk -F "_\\\[" '{print length($2)}')
let numUnderscores=32-$lenOfFirstBit
for ((j=0; $j<$numUnderscores; j=$j+1)); do
    line1="_$line1"
done

line1="# $line1"
line2="# ________________________________$line2"
# this: "${#string}"  gets the string length
let numUnderscores=${#line1}-${#line2}
for ((j=0; $j<$numUnderscores; j=$j+1)); do
    line2="${line2}_"
done

echo $line1 >> $outFile
echo "#" >> $outFile
echo $line2 >> $outFile
#Now the actual configuration
echo "" >> $outFile
echo "CFG_MOBO=$matchFile" >> $outFile
echo "MOTHERBOARD=$mbModel" >> $outFile
echo "MOTHERBOARD_SIG=$mobosig" >> $outFile
diskMapFile="${diskMapDir}/${mbModel}-diskmap.cfg"
if [ -e $diskMapFile ]; then 
  echo "MOTHERBOARD_DISKMAP_CFG=${diskMapFile}" >> $outFile
fi
echo ""   >> $outFile




for num in "${mbSlotOrderedIdx[@]}"; do

  file=${outputSlotCfgFile[$num]}
  file=$(DoQuotes "$file")

  if [ "x$file" != "x" ]; then
    outputSlotModel[$num]=$(DoQuotes "${outputSlotModel[$num]}")
    echo "CFG_SLOT${num}=$file" >> $outFile
    echo "MOBO_SLOT${num}=${outputSlotModel[$num]}" >> $outFile
    echo "CARD${num}_SIG=${outputSlotSig[$num]}" >> $outFile
    echo "" >> $outFile
  fi
done

let num=0
for ports in "${mbPort[@]}"; do
  let num=$num+1
  echo "PORT_$num=$ports" >> $outFile
done

echo "" >> $outFile
let numData=num-1
let num=0
for ports in "${mbPort[@]}"; do
  let num=$num+1
  if [ $num = 1 ]; then
     echo "CONTROL_1=1" >> $outFile
     if [ $numData = 0 ]; then
         echo "SERVICE_1=1" >> $outFile
         lastData=$(echo "DATA_$numData=2")
     else
         echo "SERVICE_2=1" >> $outFile
     fi
  elif [ $num = 2 ]; then
     echo "SERVICE_1=2" >> $outFile
     lastData=$(echo "DATA_$numData=2")
  else
     let dataNum=$num-2
     echo "DATA_$dataNum=$num" >> $outFile
  fi
done
echo $lastData >> $outFile


echo ""  >> $outFile
echo "#NMI/SMI conversion support requirements" >> $outFile
if [ "$mbNmiSmiRev" != "none specified" ]; then
    echo "MIN_BIOS_REV_SMI_NMI=$mbNmiSmiRev"  >> $outFile
fi
if [ "$mbNmiSmiDate" != "none specified" ]; then
#TODO: NORMALIZE date
    echo "MIN_BIOS_DATE_SMI_NMI=$mbNmiSmiDate" >> $outFile
fi

echo "" >> $outFile
echo "#Miscellaneous Tags" >> $outFile
for i in "${mbMisc[@]}"; do
    echo "$i" >> $outFile
done

for i in "${expMisc[@]}"; do
    echo "$i" >> $outFile
done

if [ -n "$ifRenameFile" ]; then
    echo "# Interface renames from $myAppName" > $ifRenameFile
fi
for i in "${mbRenameIf[@]}"; do
    fromName=$(echo $i | sed 's/=.*//')
    toName=$(echo $i | sed 's/.*=//')
    eval ifconfig $fromName name $toName >/dev/null 2>&1
    if [ -n "$ifRenameFile" ]; then
	echo "$fromName=$toName" >> $ifRenameFile
    fi
done

for i in "${expRenameIf[@]}"; do
    fromName=$(echo $i | sed 's/=.*//')
    toName=$(echo $i | sed 's/.*=//')
    eval ifconfig $fromName name $toName >/dev/null 2>&1
    if [ -n "$ifRenameFile" ]; then
	echo "$fromName=$toName" >> $ifRenameFile
    fi
done

for i in "${mbHideIf[@]}"; do
    eval ifconfig $i hidden >/dev/null 2>&1
done

for i in "${expHideIf[@]}"; do
    eval ifconfig $i hidden >/dev/null 2>&1
done

LogNotice "Generated motherboard mapping"
