#matchCardToCfgFile.awk
#
#This script parses the specified configuration file and determines whether
# the card is a match.

#Field separator is "="
#Start by setting the return var to 0. This way if there's an error in the
# script and we don't end up running our END section, the caller will know.
BEGIN {
    print "crdCfgFileMatch=0"
    matched = 1
    FS = "="
    firstTime = 1
    logPrefix = "mobomap: "
    matchTagsRecognized = 0
}
#Set our output variables
END {
    if (matchTagsRecognized != 0) 
    {
        if ( matched == 1 )
        {
            print "crdCfgFileMatch=1"
        }
        else
        {
            print "crdCfgFileMatch=2"
            printf ("failedTag=\"%s\"\n", failedTag)
        }
    }
    else
    {
        system ("logger -p local0.warning \"" logPrefix "(" argFileName ") No match tags specified; ignoring file\"")
        print "crdCfgFileMatch=2"
        printf ("failedTag=\"No tags found\"\n")
    }
}

function CheckForDupes( field1, field2 )
{
    if (tags[field1] == 1) 
    {
        system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Duplicate tag found (" field1 "=" field2 ")\"")
    }
    tags[field1]=1
    return 0
}

#____________________________________________________________________
#Given the pciconf info for all of the devices on the card, this function 
# checks if the line is a pciconf tag and matches the value if it is.
# Returns 1 for match, 2 for no failed match, and 0 for tag not found
#____________________________________________________________________
function MatchPciconfTag( aIdx, aDevs, aBridge, aBusses, aDevices, aFns, aPciconf, field1, field2 )
{
    offset = match( field1, /^[ \t]*match\.pciconf\./ )
    if ( offset != 0 )
    {
        matchTagsRecognized++
        CheckForDupes( field1, field2 )

        #Found a pciconf match tag, pull out the tag and value (description)
        len = offset + RLENGTH
        tag = substr( field1, len )
        
        #convert " to \"
        pattern = "\""
        if (match (field2, pattern))
        {
            str1 = substr( field2, 1, RSTART-1 )
            str2 = substr( field2, RSTART+1 )
            field2 = str1 "\\\""
            while (match (str2, pattern))
            {
                str1 = substr( str2, 1, RSTART-1 )
                str2 = substr( str2, RSTART+1 )
                field2 = field2 str1 "\\\""
            }
            field2 = field2 str2 
        }
        
        #strip off the trailing newline
        if (match (field2, /\n$/)) 
        {
            len = length ( field2 ) - 1
            description = substr( field2, 1, len )
        }
        else
        {
            description = field2
        }
        
        if ( match(tag, /^name\.[0-9]+/) )
        {
            #it's a name tag, pull out the index portion
            len = RLENGTH
            offset = match( tag, /^name\./ )
            len = len - RSTART - RLENGTH + 1
            start = RSTART + RLENGTH
            idx = substr( tag, start, len )
            #if the element in the devices array at index matches the 
            # description (the value of the tag from the config file), then
            # return 1, otherwise return 2 and set the failedTag
            if ( match( aDevs[idx], description ) )
            {
                return 1
            }
            else
            {
                failedTag = field1 "=" field2
                return 2
            }

        }
        else
        {
            #Split the pciconf info by spaces (each field is now a four byte
            # hex value prefixed with "0x")
            split ( aPciconf[idx], pciconfs, " " )
            if ( match(tag, /^cardid\.[0-9]+/) )
            {
                #It's a cardid tag.  Card ID is the 12th 4 byte value in the
                # config space.
                if ( match (description, pciconfs[12]) ) 
                { 
                    return 1
                }
                else 
                {
                    failedTag = field1 "=" field2
                    return 2 
                }
            }
            else if ( match (tag, /^classid\.[0-9]+/) )
            {
                #It's a classid tag.  Class ID is the 3rd 4 byte value in the
                # config space.
                if ( match (description, pciconfs[3]) ) 
                { 
                    return 1 
                }
                else 
                { 
                    failedTag = field1 "=" field2
                    return 2
                }
            }
            else if ( match (tag, /^chipid\.[0-9]+/) )
            {
                #It's a chipid tag.  Chip ID is the 1st 4 byte value in the
                # config space.
                if ( match (description, pciconfs[1]) ) 
                { 
                    return 1 
                }
                else 
                { 
                    failedTag = field1 "=" field2
                    return 2 
                }
            }
            else if ( match (tag, /^offset[0-9]+\.[0-9]+/) )
            {
                #It's a config space tag at an arbitrary offset. Extract the
                # offset
                offsetStart = match(tag, /[0-9]/)
                offsetLen = match( tag, /\./ ) - offsetStart
                offset = substr (tag, offsetStart, offsetLen)
                #calculate which block it will be in
                block = int(offset/4) + 1
                #offset is the character to start looking at
                offset = 2*(offset - 4*(block - 1))
                
                #if the tag's value as spec'd in the config file starts with 0x
                if (match (description, /^0x/) )
                {
                    #strip off the 0x portion
                    description = substr( description, 3 )
                }
                #Match the tag
                if ( match (substr (pciconfs[block], offset+1, length (description)), description) ) 
                { 
                    return 1 
                }
                else 
                { 
                    failedTag = field1 "=" field2
                    return 2 
                }
            }
            else
            {
                system ("logger -p local0.error \"" logPrefix "(" argFileName ") Invalid Tag: " field1 "=" field2 "\"")
                failedTag = field1 "=" field2
                return 2
            }
        }
        
    }
    return 0
}
#____________________________________________________________________
# This script attempts to (as quickly as possible) determine whether
# the specified config file matches the current card (as defined by the 
# collection of devices, pci select info, etc.).
#
# Input parameters are the number of devices and several arrays, passed in as
# newline separated records:  
#   argNumDevs - the number of devices found in this slot
#   argIdx - the indices
#   argDevs - the names of the devices (i.e. em0)
#   argBridge - the bools that inidicate whether the device is a bridge
#   argBusses - the bus numbers of the devices (i.e. 3 in pci3:1:0)
#   argDevices - the device numbers of the devices (i.e. 1 in pci3:1:0)
#   argFns - the functions of the devices (i.e. 0 in pci3:1:0) 
#   argPciconf - the pci headers (64 bytes) from pci conf, as 16 space-
#           separated 4 byte hex values (i.e. 0x12345678 0x00000000 .....)
#   argFileName - Name of the file being parsed
#
# Output parameters:
# It sets the variable crdCfgFileMatched to 1 if the specified file was
# matched, 2 if not, and 0 if an error occured in the script.
# If it fails, the tag that failed to match is written to failedTag
#____________________________________________________________________
{
    if (firstTime) 
    {
        firstTime = 0
        #arrays were passed in as newline separated value tags...parse'em
        split ( argIdx, argIdxVal, "\n" )
        split ( argDevs, argDevsVal, "\n" )
        split ( argBridge, argBridgeVal, "\n" )
        split ( argBusses, argBussesVal, "\n" )
        split ( argDevices, argDevicesVal, "\n" )
        split ( argFns, argFnsVal, "\n" )
        split ( argPciconf, argPciconfVal, "\n" )
        for (i=1; i<=argNumDevs+1; ++i) 
        {
            debug1 = i
            aIdx[i] = argIdxVal[i+1]
            aDevs[i] = argDevsVal[i+1]
            aBridge[i] = argBridgeVal[i+1]
            aBusses[i] = argBussesVal[i+1]
            aDevices[i] = argDevicesVal[i+1]
            aFns[i] = argFnsVal[i+1]
            aPciconf[i] = argPciconfVal[i+1]
        }
    }

    #if its not a comment line (does not start with #) (regex to positively match: ^[ \t]*[^#].*)
    if (match ($1, /^[ \t]*#/) == 0)
    {
        #Normalize the fields.  We expect a tag containing no '=',
        # followed by any value string.  Since the value could contain 
        # an '=' and we have to check that there are only 2 fields
        # found in this record, and if there are more, concatenate
        # $2 with the additional ones (inserting a '=')
        if (length ($1) != 0) 
        {
            if (length ($2) == 0) 
            {
                system ("logger -p local0.error \"" logPrefix "(" argFileName ") Tag missing value (" $1 ")\"")
                system ("logger -p local0.error \"" logPrefix "(" argFileName ") Skipping file\"")
                matched = 2
                exit
            }
        }
        field2 = $2
        if ( NF > 2 )
        {
            for (i=3; i<=NF; ++i)
            {
               field2 = field2 "=" $i
            }
        }
        #If it's a match tag
        if ( match( $1, /^[ \t]*match\./ ) ) 
        {
            matchedTag = MatchPciconfTag( aIdx, aDevs, aBridge, aBusses, aDevices, aFns, aPciconf, $1, field2 )            
            if ( matchedTag == 2 )
            {
                matched = 0
                exit
            }
            if (matchedTag == 0) \
            {
                system ("logger -p local0.warning \"" logPrefix "(" argFileName ") Unrecognized tag found (" $1 ")\"")
            }
        }
    }
}
