#!/bin/sh

debug()
{
    if [ $DEBUG ] ; then
	echo "$@"
    fi
}

remove_fstab_nodev()
{
    if grep -w nodev /etc/fstab > /dev/null ; then
	debug "- removing nodev option from /etc/fstab"
	sed -i .upg -e 's/,?*nodev|nodev,?//' /etc/fstab
    fi
}

add_fstab_procfs()
{
    if ! egrep -w '^proc' /etc/fstab > /dev/null ; then
	debug "- adding procfs to /etc/fstab"
	echo "proc            /proc                   "\
	    "procfs  rw                      0 0" >> /etc/fstab
    fi
}

add_fstab_noatime()
{
    if ! egrep '^/d2.*noatime' /etc/fstab > /dev/null ; then
	cp -p /etc/fstab /etc/fstab.noatime
	awk 'BEGIN {}
	{
	    if (index($2,"/d2") &&
		(! index($4,"noatime")))
            {
		gsub($4,$4",noatime")
	    }
	    print $0
	}
	END {}' < /etc/fstab.noatime > /etc/fstab
	rm /etc/fstab.noatime
    fi
}

update_master_passwd()
{
    local ifs name pw uid gid home shell mflag
    ifs="$IFS"
    IFS=:
    while read name pw uid gid desc home shell ; do
	if pw usershow -n "$name" >/dev/null 2>/dev/null ; then
	    # User exists, make sure it is correct.
	    curhome="`pw usershow $name 2>/dev/null | awk -F : '{print $9}'`"
	    if [ "$curhome" != "$home" ] ; then
		case "$home" in
		    /home/*)
			mflag="-m"
			;;
		esac
		pw usermod $name -d "$home" $mflag
		debug "- user $name homedir updated"
	    else
		debug "- user $name OK"
	    fi
	else
	    # User needs to be added.
	    if ! pw groupshow -n $gid >/dev/null 2>/dev/null ; then
		pw groupadd "$name" $gid
		debug "- group $name added with gid $gid"
	    fi
	    pw useradd "$name" -u $uid -c "$desc" -d "$home" -s "$shell"
	    debug "- user $name added with uid $uid"
	fi
    done <<EOF
proxy:*:62:62:Packet Filter pseudo-user:/nonexistent:/usr/sbin/nologin
_pflogd:*:64:64:pflogd privsep user:/var/empty:/usr/sbin/nologin
_dhcp:*:65:65:dhcp programs:/var/empty:/usr/sbin/nologin
sv:*:196:198:User &:/home/sv:/bin/bash
sv_admin:*:197:199:User &:/home/sv_admin:/bin/bash
sv_service:*:198:199:User &:/home/sv_service:/bin/bash
sv_operator:*:199:199:User &:/home/sv_operator:/bin/bash
EOF
    IFS="$ifs"
    chpass -p \* toor 2>/dev/null
}

set_permissions()
{
    chmod 700 /root
    chown root:wheel /root

    # 5.00.0039
    chmod 644 /etc/dhclient-exit-hooks 2>/dev/null
}

install_boot_block()
{
    disklabel -B `sysctl -n kern.rootdev`
}

cleanup_files()
{
    rm -f /etc/pam.conf 2>/dev/null
    rm -f /usr/share/man/man8/lmmon.8.gz 2>/dev/null
    rm -f /root/.exrc 2>/dev/null
    # Clean out preformatted man page cache (reqst00034005)
    find /usr/share/man -type f -path '/usr/share/man/cat*/*' -delete
}

clean_motd()
{
    # reqst00078279: Old upgrade packages used to update /etc/motd with the
    # image version number.  Remove these.
    sed -i '' -E '/upgraded to [4-7]\.[0-9]{2}\.[0-9]{4} on [0-9]{8}-[0-9]{2}:[0-9]{2}:[0-9]{2}/d' /etc/motd
}

clean_rc_conf_local()
{
    # reqst00080208: This line often ends up in rc.conf.local (usually by
    # someone creating rc.conf.local as a copy of another file).  Remove it as
    # it causes syslog-ng to not start.
    sed -i '' -E '/^syslogd_flags="-ss"/d' /etc/rc.conf.local
}

configure_hosts_allow()
{
    if [ ! -f /etc/hosts.allow ]; then
	return 0
    fi
    # reqst00099184 uncomment localhost
    sed -i '' -e 's/^#ALL : localhost.*$/ALL : localhost 127.0.0.1 : allow/' \
	/etc/hosts.allow
}

rm_files_filelist()
{
    while read file; do
	rm -f $file
    done < $1
}

rm_dirs_filelist()
{
    while read dir; do
	rmdir $dir 2>/dev/null
    done < $1
}

rm_arch_old_files_dirs()
{
    if [ -f $UPGDIR/old_files.$ARCH ]; then
	rm_files_filelist $UPGDIR/old_files.$ARCH
    fi
    if [ -f $UPGDIR/old_dirs.$ARCH ]; then
	rm_dirs_filelist $UPGDIR/old_dirs.$ARCH
    fi
}

# reqst00157946 - "obsolete" files and dirs are those that we have decided to
# remove from the image, and should be deleted on every upgrade.
rm_obsolete_files_dirs()
{
    if [ -f $UPGDIR/obsolete_files ]; then
	rm_files_filelist $UPGDIR/obsolete_files
    fi
    if [ -f $UPGDIR/obsolete_dirs ]; then
	rm_dirs_filelist $UPGDIR/obsolete_dirs
    fi
}

# Upgrade from FreeBSD 4.x based systems (5.xx.xxxx)
upgrade_from_4x()
{
    configure_hosts_allow
    update_master_passwd
    add_fstab_procfs
    add_fstab_noatime
    cleanup_files
    rm_arch_old_files_dirs
    rm_obsolete_files_dirs
    clean_motd
    clean_rc_conf_local
    set_permissions
}
# Upgrade from FreeBSD 5.x based systems (7.xx.xxxx)
upgrade_from_5x()
{
    upgrade_from_4x
}

# Invoked as the final stage of upgrades that do not involve a major bump.
finish_upgrade()
{
    configure_hosts_allow
    update_master_passwd
    rm_obsolete_files_dirs
}

ARCH=`uname -m`
UPGDIR=/usr/local/sandvine/upgrade/

while getopts "dx" Option ; do
    case $Option in
	d)
	    DEBUG=1
	    ;;
	x)
	    set -x
	    ;;
	*)
	    exit 1
	    ;;
    esac
done
shift `expr $OPTIND - 1`

if [ $# -eq 0 ] ; then
    echo "Usage: `basename $0` [options] <cmd> [<cmd>...]"
    echo
    exit 1
fi

for cmd in "$@" ; do
    debug "Invoking \"$cmd\""
    eval "$cmd"
    debug ""
done
