#!/bin/sh

# -xv will turn on verbose debugging
# Want to set up a swap partition that is > than hw.physmem
# Want to set up a single large partition (/d2) that is the
# remainder of the disk, so that the layout is
# / swap /d2
getDiskCnt()
{
    buf=$(sysctl -n kern.disks)
    if [ $? -ne 0 ]
    then
        echo "Failure on sysctl -n kern.disks"
        exit 2
    fi
    containers=$(echo $buf | 
        awk 'BEGIN {cnt=0}
        {
            for(i=1; i<=NF; i++) 
            {
                if (!match($i,"md[0-9]+"))
                {
                    ++cnt
                }
            }
        } END {print cnt}')
    echo $containers
}

getDiskList()
{
    buf=$(sysctl -n kern.disks)
    if [ $? -ne 0 ]
    then
        echo "Failure on sysctl -n kern.disks"
        exit 3
    fi

    disks=$( echo $buf | 
        awk '
        {
            for(i=1; i<=NF; i++) 
            {
                if (!match($i,"md[0-9]+"))
                {
                    print $i
                }
            }
        }')
    echo $disks
}

# $1=disk name and $2=partition number without the colon
getDiskSizePartX ()
{
	local lsize
	lsize=`fdisk -s $1 | awk "/^ *$2:/ { print \\$3 }"`
	if [ -z "$lsize" ]; then
		lsize=0
	fi
	echo $lsize
}

# Get required swap partition size, calculated as memory rounded up to 1GB
# boundary plus some extra for crash dump headers.  Value returned in MB.
getSwapSize()
{
    local GB mem mem_gb padding
    padding=32

    mem=`sysctl -n hw.physmem`
    if [ $? -ne 0 ]; then
        echo "Failure getting hw.physmem"
        exit 2
    fi

    GB=1073741824

    # Round up to the next integer GB, return in MB with padding.
    echo $mem $GB $padding | \
        awk 'END {print int(($1 + $2 - 1) / $2) * 1024 + $3}'
}

doRootDisk() 
{
    # Create a root and swap partition from the first container
    # the swap partition has to be greater than the hw.physmem

    if ! grep -q swap /etc/fstab 
    then
	echo "Create d2, swap data"
	root_part=$(df / |grep -v Filesystem | awk '{print $1}')
	root_disk=$(echo $root_part | sed -e 's?/dev/??' -e 's?s1a??g')

	if [ -z "$root_disk" ]
	then
	    echo "Can't figure out which disk is the root disk"
	    exit 2
	fi

	containerCnt=$(getDiskCnt)

	if [ $containerCnt -eq 0 ]
	then
	    echo "No containers found exiting"
	    exit 2
	elif [ $containerCnt -ge 3 ]
	then
	    echo "Number of disks $containerCnt is greater than expected (2)"
	fi    

	containerList=$(getDiskList)
	
	slice="${root_disk}s2"
	swap=/dev/${slice}b

	ssize=$(getSwapSize)
	psize=$(getDiskSizePartX $root_disk 2 )

	if [ $psize -lt $ssize ]
	then
	    echo "partition size ($psize) for $root_disk 2: is smaller then the physical mem size $ssize" 
	    exit 2
	fi

	echo "disklabel and newfs second partion:"
	# start from fresh for all O/S/ versions
	echo "Creating swap and /d2 on first container"

	disklabel -w -r ${slice} auto
	(
	    disklabel ${slice} | grep -v -e "^ *a: "
	    echo "b: ${ssize}m 0 swap"
	    echo "e: * * 4.2BSD"
	) | disklabel -R ${slice} /dev/stdin
        if [ $? -ne 0 ]
        then
            echo "Can't disklabel second partition"
            exit 2
        fi

	swapSlice=${slice}b
	d2slice=${slice}e

	swapon $swap
	dumpon $swap
        
	awk -v RFS=$root_part '
	{
	    if ($2 != "/" || $1 == RFS)
		printf("%s\n", $0);
	}' < /etc/fstab >/etc/fstab.new
	mv /etc/fstab.new /etc/fstab
        sed -i '' -e '/^vfs.root.mountfrom=.*# setup_fstab/d' /boot/loader.conf

	newfs -U -b 16384 -f 2048 -i 16384 /dev/$d2slice >/dev/null
	if [ $? -ne 0 ]
	then
	    echo "Can't newfs the d2 partition on $d2slice"
	    exit 2
	fi
	(
	    echo "/dev/$d2slice	/d2			ufs	rw,noatime		2 2"
	    echo "/dev/$swapSlice	none			swap	sw			0 0"
	) >> /etc/fstab

	mkdir -p /d2
	mount /d2
	if [ $? -ne 0 ]
	then
	    echo "Can't mount /d2"
	    exit 2
	fi

	mkdir /d2/tmp
	mkdir -p /d2/local
	cp -pR /usr/local/* /d2/local
	chmod u+t,a+rwx /d2/tmp
	rm -rf -- /var/tmp
	ln -sf ../d2/tmp /var/tmp
	rm -rf /usr/local
	ln -s  ../d2/local /usr/local
	mkdir -p /d2/crash
	cp -pR /var/crash/* /d2/crash
	rm -rf /var/crash
	ln -s ../d2/crash /var/crash
    fi #if ! grep -q swap /etc/fstab
}

# Check that the disk named by the expr(1) expression $1 is in /etc/fstab
check_fstab()
{
	while read dev mp type more; do
		if expr "$dev" : "/dev/$1" >/dev/null; then
			return 0
		fi
	done < /etc/fstab
	return 1
}

# Check that the disk named by the expr(1) expression in $1 is in the current
# mount list 
check_mount()
{
	/sbin/mount |
	(
		while read dev on mntpt more; do
			if expr "$dev" : "/dev/$1" >/dev/null; then
				return 0
			fi
		done
		return 1
	)
}

doAllNonRootDisks() 
{
    # First, find all the disks (other than memory devices) in the 
    # system.
    disks=$(sysctl kern.disks | 
        awk '
        {
            numd = 0
            delete disks
            for(i=2; i<=NF; i++) 
            {
                if (!match($i,"md[0-9]+"))
                {
                    disks[substr($i,match($i,"[0-9]"))]=$i
                    numd++
                }
            }
        }
        END {
            for (d=1; d < numd; d++)
            {
                print disks[d]
            }
        }')

    fsnum=3
    for d in $disks
    do
	fs="/d${fsnum}"
	fsnum=$(($fsnum + 1))
	if check_fstab "$d"; then
	    echo "drive $d already setup"
	    continue
	fi
	if check_mount "$d"; then
	    echo "Error: drive $d not in fstab, but is mounted"
	    continue
	fi
	# reqst00140635: Clear any existing fdisk partition table
	dd if=/dev/zero of=/dev/$d bs=512 count=1 >/dev/null
	gpt destroy $d 2>/dev/null
	if ! gpt create -f $d || ! gpt add -t ufs $d; then
	    echo "Error creating gpt on $d"
	    exit 2
	fi
	newfs -U -b 16384 -f 2048 -i 262144 -g 262144 /dev/${d}p1 >/dev/null
        if [ $? -ne 0 ]
        then
            echo "Can't newfs drive: ${d}p1"
            exit 2
        fi
	echo "/dev/${d}p1	$fs			ufs	rw,noatime		2 2" \
	    >> /etc/fstab
	mkdir -p $fs
	mount $fs
    done
}

log()
{
	logger "$@"
	echo "$@" >&2
}

move_and_symlink()
{
	# If the source is a symlink then we've already run.
	if [ -h $1 ] ; then
		return 0
	fi

	mkdir -p $2

	_cursize=`du -ks $1 | awk '{print $1}'`
	_freespace=`df -kn $2 | awk 'NR>1 {print $4}'`

	# Arbitrarily require at least 10MB left on the destination.
	if [ $(($_freespace - $_cursize)) -lt 10240 ] ; then
		log "Insufficient space to move $1 to $2"
		return 1
	fi

	tar -c -C $1 -f - . | tar -C $2 -x -p -f -
	if [ $? -ne 0 ] ; then
		log "Error copying $1 to $2"
		rm -rf $2
		return 1
	fi

	rm -rf $1
	ln -fs $2 $1
}

stop_syslog()
{
	_i=1
	_pidfile=/var/run/syslog.pid
	_pid=`cat $_pidfile 2>/dev/null`
	if [ $? -ne 0 ] ; then
		log "Syslog pidfile not found.  Assuming not running."
		return 0
	fi
	kill $_pid 2>/dev/null
	while [ $_i -le 10 ] ; do
		if ! kill -0 $_pid 2>/dev/null ; then
			break
		fi
		kill $_pid 2>/dev/null
		sleep 2
	done
	kill -9 $_pid 2>/dev/null
	rm -f $_pidfile
}

start_syslog()
{
	/etc/rc.d/syslogd start
}

setup_d2()
{
	if [ ! -h /var/log ] ; then
		if [ -d /d2/log ] ; then
			log "/d2/log already exists - /var/log not moved."
		else
			stop_syslog
			move_and_symlink /var/log /d2/log
			start_syslog
		fi
	fi
}

case $1 in
start)
	    usingKernVer=0
            version=`uname -b 2>/dev/null | awk -F . '{print $1}'`
            if [ $? -ne 0 ]
            then
                usingKernVer=1
                version=`uname -r | awk -F . '{print $1}'`
                if [ $? -ne 0 ]
                then
                    version=4
                    echo "Unable to determine system version using default $version" 
                fi
            fi                
            # first, do the root disk special, it gets the swap 
	    doRootDisk
	    # Now do the remainder
	    doAllNonRootDisks
	    setup_d2
	;;

stop)
	;;

restart)
	;;

*)
	echo "usage: `basename $0` {start|stop|restart}" >&2
	exit 64
	;;
esac

exit 0
